# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import os
import sys
sys.path.insert(0, os.path.abspath('..'))     # for autodoc

def get_version_string(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(here, rel_path), 'r') as fp:
        for line in fp:
            if line.startswith('__version__'):
                delim = '"' if '"' in line else "'"
                return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")

__version__ = get_version_string('../pyslsqp/__init__.py')

# -- Project information -----------------------------------------------------

project = 'PySLSQP'
copyright = '2024, Anugrah'
author = 'Anugrah'
version = __version__
release = __version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx_rtd_theme",
    # "sphinx.ext.autodoc",           # for auto generating API reference
    "numpydoc",                 
    "sphinx_copybutton",            # allows copying code embedded in the docs rendered from .md or .ipynb files
    "myst_nb",                      # renders .md, .myst, .ipynb files
    "sphinx.ext.viewcode",          # adds the source code for classes and functions in auto generated api ref
    "sphinxcontrib.bibtex",         # for references and citations
]

autodoc_typehints = 'none'
autosummary_generate = True
# add_module_names = False

# sphinxcontrib.bibtex options
bibtex_bibfiles = ['src/references.bib']

# myst_options
myst_heading_anchors = 3 # To use heading anchors for all headings of level 3 and above

# myst_nb options
myst_title_to_header = True
myst_enable_extensions = ["dollarmath", "amsmath", "tasklist"]
nb_execution_mode = 'off'

# napoleon options
# napoleon_use_ivar = False
# napoleon_use_param = False

root_doc = 'index'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['README.md', '_build', 'Thumbs.db', '.DS_Store', 'src/welcome.md']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = 'sphinx_rtd_theme' # other theme options: 'sphinx_book_theme', 'sphinx_rtd_theme', 
                                # 'alabaster', 'classic', 'sphinxdoc', 'nature', 'bizstyle', ...

# html_theme_options for sphinx_rtd_theme
html_theme_options = {
    'logo_only': False,
    'display_version': True,
    'prev_next_buttons_location': 'bottom',
    'style_external_links': False,
    'vcs_pageview_mode': '',
    'style_nav_header_background': '#2980B9',   # other valid colors: 'white', ...
    # toc options
    'collapse_navigation': False,   # default: True
    'sticky_navigation': True,
    'navigation_depth': 4,
    'includehidden': True,
    'titles_only': True     # default: False
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']


import warnings
import numpy as np
try:
    import h5py
except ImportError:
    warnings.warn("h5py not found, saving and loading data disabled")
    h5py = None

def import_h5py_file(filepath):
    if h5py is None:
        raise ImportError("h5py not found, saving and loading data disabled")
    try:
        return h5py.File(filepath, 'r')
    except:
        raise FileNotFoundError(f"File {filepath} not found or not a valid h5py file.")
    
def save_iteration(file, iter, vars, out_dict):
    '''
    Save the data from one iteration to an active file.

    Parameters
    ----------
    file : str
        Loaded file.
    iter : int
        Iteration number.
    vars : list
        List of variable names to save.
    out_dict : dict
        Dictionary with variable names as keys and variable values as values.
    '''
    file.create_group('iter_' + str(iter))
    for var in vars:
        file[f'iter_{iter}'][var] = out_dict[var]
    
def print_dict_as_table(data):
    """
    Print any input dictionary as a table.

    Parameters
    ----------
    data : dict
        Dictionary to print as a table.

    Examples
    --------
    >>> data = {'a': 0, 'b': "string", 'c': ['a', 'b', 'c']}
    >>> print_dict_as_table(data)
    --------------------------------------------------
            a                        : 0
            b                        : string
            c                        : ['a', 'b', 'c']
    --------------------------------------------------
    """
    print("--------------------------------------------------")
    for key, value in data.items():
        print(f"        {key:24} : {value}")
    print("--------------------------------------------------")
    
def print_file_contents(filepath):
    '''
    Print the contents of the saved file.
    
    Parameters
    ----------
    filepath : str
        Path to the saved file.        

    Examples
    --------
    >>> import numpy as np
    >>> from pyslsqp import optimize
    >>> obj = lambda x: np.sum(x**2)
    >>> grad = lambda x: 2*x
    >>> xl = 0.0
    >>> xu = np.array([1, 1])
    >>> x0 = np.array([0.5, 0.5])
    >>> results = optimize(x0, obj=obj, grad=grad, xl=xl, xu=xu, save_itr='major', save_vars=['objective', 'optimality', 'x'])  # doctest: +ELLIPSIS
    No constraints defined. Running an unconstrained optimization problem...
    Optimization terminated successfully    (Exit mode 0)
                Final objective value                : 0.000000e+00
                Final optimality                     : 0.000000e+00
                Final feasibility                    : 0.000000e+00
                Number of major iterations           : 2
                Number of function evaluations       : 2
                Number of derivative evaluations     : 2
                Average Derivative evaluation time   : ... s per evaluation
                Average Function evaluation time     : ... s per evaluation
                Total Function evaluation time       : ... s [ ...%]
                Total Derivative evaluation time     : ... s [ ...%]
                Optimizer time                       : ... s [ ...%]
                Processing time                      : ... s [ ...%]
                Visualization time                   : ... s [  0.00%]
                Total optimization time              : ... s [100.00%]
                Summary saved to                     : slsqp_summary.out
                Iteration data saved to              : slsqp_recorder.hdf5
    >>> from pyslsqp.postprocessing import print_file_contents
    >>> print_file_contents('slsqp_recorder.hdf5')  # doctest: +NORMALIZE_WHITESPACE
    Available data in the file:
    ---------------------------
         Attributes of optimization  : ['acc', 'con_scaler', 'finite_diff_abs_step', 'finite_diff_rel_step', 
         'hot_start', 'iprint', 'keep_plot_open', 'load_filename', 'm', 'maxiter', 'meq', 'n', 'obj_scaler', 
         'save_figname', 'save_filename', 'save_itr', 'save_vars', 'summary_filename', 'visualize', 'visualize_vars', 
         'warm_start', 'x0', 'x_scaler', 'xl', 'xu']
         Saved variable iterates     : ['ismajor', 'iter', 'majiter', 'objective', 'optimality', 'x']
         Results of Optimization     : ['constraints', 'feasibility', 'fev_time', 'gev_time', 'gradient', 'jacobian', 
         'message', 'multipliers', 'nfev', 'ngev', 'num_majiter', 'objective', 'optimality', 'optimizer_time', 'processing_time', 
         'save_filename', 'status', 'success', 'summary_filename', 'total_time', 'visualization_time', 'x']

    '''
    file = import_h5py_file(filepath)
    print("Available data in the file:")
    print("---------------------------")
    try:
        print("     Attributes of optimization  :", list(file.attrs.keys()))
    except:
        warnings.warn("No attributes found in the file.")
    try:
        print("     Saved variable iterates     :", list(file['iter_0'].keys()))
    except:
        warnings.warn("No variable iterates found in the file.")
    try:
        print("     Results of Optimization     :", list(file['results'].keys()))
    except:
        warnings.warn("No results found in the file.")
    file.close()

def load_variables(filepath, vars, itr_start=0, itr_end=-1, major_only=False):
    '''
    Load specified variable iterates between ``itr_start`` and ``itr_end`` from the saved file.
    Returns a dictionary with the variable names as keys and list of variable iterates as values.
    Note the variables at ``itr_start`` and ``itr_end`` are included in the output.

    Parameters
    ----------
    filepath : str
        Path to the saved file.
    vars : str or list
        Variable names to load from the saved file.
    itr_start : int, default=0
        Starting iteration to load the variables from.
        Negative indices are allowed with -1 representing the last iteration
        and -2 representing the second last iteration and so on.
    itr_end : int, default=-1
        Ending iteration to load the variables from. 
        Negative indices are allowed with -1 representing the last iteration
        and -2 representing the second last iteration and so on.
    major_only : bool, default=False
        If True, only major iterations are loaded.
        If False, all iterations are loaded irrespective of major or line search iterations.

    Returns
    -------
    out_data : dict
        Dictionary with variable names as keys and list of variable iterates as values.

    Examples
    --------
    >>> import numpy as np
    >>> from pyslsqp import optimize
    >>> obj = lambda x: np.sum(x**2)
    >>> grad = lambda x: 2*x
    >>> xl = 0.0
    >>> xu = np.array([1, 1])
    >>> x0 = np.array([0.5, 0.5])
    >>> results = optimize(x0, obj=obj, grad=grad, xl=xl, xu=xu, save_itr='major', save_vars=['objective', 'optimality', 'x'])  # doctest: +ELLIPSIS
    No constraints defined. Running an unconstrained optimization problem...
    Optimization terminated successfully    (Exit mode 0)
                Final objective value                : 0.000000e+00
                Final optimality                     : 0.000000e+00
                Final feasibility                    : 0.000000e+00
                Number of major iterations           : 2
                Number of function evaluations       : 2
                Number of derivative evaluations     : 2
                Average Derivative evaluation time   : ... s per evaluation
                Average Function evaluation time     : ... s per evaluation
                Total Function evaluation time       : ... s [ ...%]
                Total Derivative evaluation time     : ... s [ ...%]
                Optimizer time                       : ... s [ ...%]
                Processing time                      : ... s [ ...%]
                Visualization time                   : ... s [  0.00%]
                Total optimization time              : ... s [100.00%]
                Summary saved to                     : slsqp_summary.out
                Iteration data saved to              : slsqp_recorder.hdf5
    >>> from pyslsqp.postprocessing import load_variables
    >>> load_variables('slsqp_recorder.hdf5', ['objective', 'optimality', 'x[0]'], itr_start=0, itr_end=-1, major_only=True)
    {'objective': [0.5, 0.0, 0.0], 'optimality': [99.0, 0.0, 0.0], 'x[0]': [0.5, 0.0, 0.0]}

    '''
    if not isinstance(filepath, str):
        raise ValueError("filepath must be a string.")
    
    if not isinstance(vars, (str, list)):
        raise ValueError("vars must be a string or a list of strings")
    if isinstance(vars, str):
        vars = [vars]
    if not all(isinstance(var, str) for var in vars):
        raise ValueError("vars must be a string or a list of strings")
    
    if not isinstance(itr_start, int):
        raise ValueError("itr_start must be an integer.")
    if not isinstance(itr_end, int):
        raise ValueError("itr_end must be an integer.")
    
    file  = import_h5py_file(filepath)
    niter = len(file.keys()) - 2 # Number of iterations saved in the file [excludes 0th iteration and results]
    num_saves = len(file.keys()) - 1 # Number of iterations saved [includes 0th iteration but excludes results]
    if major_only:
        niter = file['results']['num_majiter'][()]
    if (-(niter+1) <= itr_start <= niter) and (-(niter+1) <= itr_end <= niter):
        start = itr_start * 1
        if itr_start < 0:
            start = niter + itr_start + 1
        end = itr_end * 1
        if itr_end < 0:
            end = niter + itr_end + 1
        if start > end:
            raise ValueError(f"itr_start index ({start}) must be less than itr_end index ({end}).")
    else:
        raise ValueError(f"itr_start {itr_start} and itr_end {itr_end} must be within bounds (>={-(niter+1)} and <={niter}).")
    

    out_data = {}
    for var in vars:
        if var.split('[')[0] not in file['iter_0'].keys():
            raise ValueError(f"Variable {var} not found in the file.")
        
        out_data[var] = []

    if major_only:
        for i in range(num_saves):
            if file[f'iter_{i}']['ismajor'][()] and file[f'iter_{i}']['majiter'][()] >= start:
                for var in vars:
                    if '[' in var:
                        name = var.split('[')[0]
                        if name == 'jacobian':
                            idx1, idx2 = map(int, var.split('[')[1].split(']')[0].split(','))
                            out_data[var].append(file[f'iter_{i}'][name][idx1, idx2])
                        else:
                            idx = int(var.split('[')[1].split(']')[0])
                            out_data[var].append(file[f'iter_{i}'][name][idx])
                    else:
                        out_data[var].append(file[f'iter_{i}'][var][()])
                if file[f'iter_{i}']['majiter'][()] == end:
                    break
    else:
        for i in range(start, end+1):
            for var in vars:
                if '[' in var:
                    name = var.split('[')[0]
                    if name == 'jacobian':
                        idx1, idx2 = map(int, var.split('[')[1].split(']')[0].split(','))
                        out_data[var].append(file[f'iter_{i}'][name][idx1, idx2])
                    else:
                        idx = int(var.split('[')[1].split(']')[0])
                        out_data[var].append(file[f'iter_{i}'][name][idx])
                else:
                    out_data[var].append(file[f'iter_{i}'][var][()])
    
    file.close()

    return out_data

def load_results(filepath):
    '''
    Load the results of optimization from the saved file as a dictionary.

    Parameters
    ----------
    filepath : str
        Path to the saved file.

    Returns
    -------
    out_data : dict
        Dictionary with optimization results.

    Examples
    --------
    >>> import numpy as np
    >>> from pyslsqp import optimize
    >>> obj = lambda x: np.sum(x**2)
    >>> grad = lambda x: 2*x
    >>> xl = 0.0
    >>> xu = np.array([1, 1])
    >>> x0 = np.array([0.5, 0.5])
    >>> results = optimize(x0, obj=obj, grad=grad, xl=xl, xu=xu, save_itr='major', save_vars=['objective', 'optimality', 'x'])  # doctest: +ELLIPSIS
    No constraints defined. Running an unconstrained optimization problem...
    Optimization terminated successfully    (Exit mode 0)
                Final objective value                : 0.000000e+00
                Final optimality                     : 0.000000e+00
                Final feasibility                    : 0.000000e+00
                Number of major iterations           : 2
                Number of function evaluations       : 2
                Number of derivative evaluations     : 2
                Average Derivative evaluation time   : ... s per evaluation
                Average Function evaluation time     : ... s per evaluation
                Total Function evaluation time       : ... s [ ...%]
                Total Derivative evaluation time     : ... s [ ...%]
                Optimizer time                       : ... s [ ...%]
                Processing time                      : ... s [ ...%]
                Visualization time                   : ... s [  0.00%]
                Total optimization time              : ... s [100.00%]
                Summary saved to                     : slsqp_summary.out
                Iteration data saved to              : slsqp_recorder.hdf5
    >>> from pyslsqp.postprocessing import load_results
    >>> load_results('slsqp_recorder.hdf5')  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    {'constraints': array([], dtype=float64), 'feasibility': 0.0, 'fev_time': ..., 'gev_time': ..., 'gradient': array([0., 0.]), 
    'jacobian': array([], shape=(0, 2), dtype=float64), 'message': 'Optimization terminated successfully', 
    'multipliers': array([], dtype=float64), 'nfev': 2, 'ngev': 2, 'num_majiter': 2, 'objective': 0.0, 'optimality': 0.0, 
    'optimizer_time': ..., 'processing_time': ..., 'save_filename': 'slsqp_recorder.hdf5', 'status': 0, 'success': True, 
    'summary_filename': 'slsqp_summary.out', 'total_time': ..., 'visualization_time': 0.0, 'x': array([0., 0.])}

    '''
    file = import_h5py_file(filepath)
    result_dict = {}
    for key in file['results'].keys():
        result_dict[key] = file['results'][key][()]
        if key in ['message', 'save_filename', 'summary_filename']:
            result_dict[key] = result_dict[key].decode('utf-8')
    file.close()
    return result_dict

def load_attributes(filepath):
    '''
    Load the attributes of optimization from the saved file as a dictionary.

    Parameters
    ----------
    filepath : str
        Path to the saved file.

    Returns
    -------
    out_data : dict
        Dictionary with optimization attributes.

    Examples
    --------
    >>> import numpy as np
    >>> from pyslsqp import optimize
    >>> obj = lambda x: np.sum(x**2)
    >>> grad = lambda x: 2*x
    >>> xl = 0.0
    >>> xu = np.array([1, 1])
    >>> x0 = np.array([0.5, 0.5])
    >>> results = optimize(x0, obj=obj, grad=grad, xl=xl, xu=xu, save_itr='major', save_vars=['objective', 'optimality', 'x'])  # doctest: +ELLIPSIS
    No constraints defined. Running an unconstrained optimization problem...
    Optimization terminated successfully    (Exit mode 0)
                Final objective value                : 0.000000e+00
                Final optimality                     : 0.000000e+00
                Final feasibility                    : 0.000000e+00
                Number of major iterations           : 2
                Number of function evaluations       : 2
                Number of derivative evaluations     : 2
                Average Derivative evaluation time   : ... s per evaluation
                Average Function evaluation time     : ... s per evaluation
                Total Function evaluation time       : ... s [ ...%]
                Total Derivative evaluation time     : ... s [ ...%]
                Optimizer time                       : ... s [ ...%]
                Processing time                      : ... s [ ...%]
                Visualization time                   : ... s [  0.00%]
                Total optimization time              : ... s [100.00%]
                Summary saved to                     : slsqp_summary.out
                Iteration data saved to              : slsqp_recorder.hdf5
    >>> from pyslsqp.postprocessing import load_attributes
    >>> load_attributes('slsqp_recorder.hdf5')  # doctest: +NORMALIZE_WHITESPACE +ELLIPSIS
    {'acc': 1e-06, 'con_scaler': 1.0, 'finite_diff_abs_step': 'None (undefined)', 'finite_diff_rel_step': 1.4901161193847656e-08, 
    'hot_start': False, 'iprint': 1, 'keep_plot_open': False, 'load_filename': 'None (undefined)', 'm': 0, 'maxiter': 100, 'meq': 0, 'n': 2, 'obj_scaler': 1.0, 
    'save_figname': 'slsqp_plot.pdf', 'save_filename': 'slsqp_recorder.hdf5', 'save_itr': 'major', 'save_vars': ['objective', 'optimality', 'x'], 
    'summary_filename': 'slsqp_summary.out', 'visualize': False, 'visualize_vars': ['objective', 'optimality', 'feasibility'], 'warm_start': False, 
    'x0': array([0.5, 0.5]), 'x_scaler': 1.0, 'xl': 0.0, 'xu': array([1, 1])}

    '''
    file = import_h5py_file(filepath)
    attr_dict = {}
    for key in file.attrs.keys():
        attr_dict[key] = file.attrs[key]
        if key in ['save_vars', 'visualize_vars']:
            attr_dict[key] = list(file.attrs[key])
    file.close()
    return attr_dict

if __name__ == "__main__":
    import doctest
    doctest.testmod()



'''
This file contains tests for the examples within docstrings.
'''
import os
import sys
here = os.path.abspath(os.path.dirname(__file__))
sys.path.append(os.path.join(here, '../pyslsqp'))  # Add the pyslsqp directory to the Python path
import doctest

def test_main():
    import main
    failures, _ = doctest.testmod(main)
    assert failures == 0, 'One or more doctests failed in main.py'

import pytest

@pytest.mark.visualize
def test_visualize():
    import visualize
    if os.getenv("GITHUB_ACTIONS") is None: # Skip this test on GitHub Actions since it requires a display
        failures, _ = doctest.testmod(visualize)
        assert failures == 0, 'One or more doctests failed in visualize.py'


def test_save_and_load():
    import save_and_load
    failures, _ = doctest.testmod(save_and_load)
    assert failures == 0, 'One or more doctests failed in save_and_load.py'

if __name__ == '__main__':
    test_main()
    test_visualize()
    test_save_and_load()


import warnings
import numpy as np
import time

try:
    import matplotlib
    matplotlib.use('TkAgg')
    import matplotlib.pyplot as plt

except ImportError:
    warnings.warn("matplotlib not found, plotting disabled.")
    plt = None

class Visualizer:

    def __init__(self, visualize_vars, summary_filename, save_figname):
        '''
        Initialize the visualizer with the variables to visualize. 
        The variables should be a list of strings, where each string is the name of a variable to visualize. 
        The variables can be any of the following:
            - 'objective'       : the objective function value
            - 'optimality'      : the optimality condition
            - 'feasibility'     : the feasibility condition
            - 'x[i]'            : the ith variable value
            - 'constraints[i]'  : the ith constraint value
            - 'jacobian[i,j]'   : the (i,j) element of the Jacobian matrix
            - 'gradient[i]'     : the ith gradient value
            - 'multipliers[i]'  : the ith Lagrange multiplier value

        Creates an interactive plot with the specified variables on the y-axis and the iteration number on the x-axis.
        The plots are stacked vertically in the order they are specified in the list.
        The plot is updated with the latest values of the variables after each iteration.

        Parameters
        ----------
        visualize_vars : list of str
            List of variables to visualize.
        summary_filename : str
            Name of the summary file which is displayed in the title of the plot.
        save_figname : str
            Name of the file to save the plot.
        '''

        v_start = time.time()
        if plt is None:
            raise ImportError("matplotlib not found, cannot visualize.")
        self.visualize_vars = visualize_vars
        self.save_figname = save_figname
        plt.ion()
        lines_dict = {}
        var_dict = {}
        n_plots = len(visualize_vars)
        self.fig, self.axs = plt.subplots(n_plots, figsize=(10, 3*n_plots))
        self.fig.suptitle(f'SLSQP Optimization [{summary_filename}]')
        for ax, var in zip(self.axs, visualize_vars):
            # ax.set_title(var)
            # ax.set_xlabel('Iteration')
            ax.set_ylabel(var)
            var_dict[var] = []
            if var in ['optimality', 'feasibility']:
                lines_dict[var], = ax.semilogy([], [], label=var)
            else:
                lines_dict[var], = ax.plot([], [], label=var)
            ax.legend()

        # self.fig.set_figwidth(8)
        # self.fig.set_figheight(3*n_plots)
        # self.fig.set_size_inches(10, 3*len(self.visualize_vars), forward=True)
        # plt.gcf().set_size_inches(10, 3*len(self.visualize_vars))
        plt.tight_layout(pad=3.0, h_pad=0.1, w_pad=0.1, rect=[0, 0, 1., 1.])

        self.lines_dict = lines_dict
        self.var_dict = var_dict

        self.vis_time = time.time() - v_start
        self.wait_time = 0.0

    def update_plot(self, out_dict):
        '''
        Update the plot with the latest values of the variables.
        Appends the values of scalar iterates after each iteration in the var_dict attribute before updating the plot.
        The out_dict should be a dictionary containing the following
        keys:
            - 'majiter'     : the number of major iterations
            - 'objective'   : the objective function value
            - 'optimality'  : the optimality condition
            - 'feasibility' : the feasibility condition
            - 'x'           : the variable values
            - 'constraints' : the constraint values
            - 'jacobian'    : the Jacobian matrix
            - 'gradient'    : the gradient values
            - 'multipliers' : the Lagrange multiplier values
        '''

        v_start = time.time()
        x_data = np.arange(out_dict['majiter']+1)
        for k, var in enumerate(self.visualize_vars):
            if var in ['objective', 'optimality', 'feasibility']:
                self.var_dict[var].append(out_dict[var]*1.0) # *1.0 is necessary so that the value is not a reference
            elif var.startswith('jacobian['):
                idx1, idx2 = map(int, var[9:-1].split(','))
                self.var_dict[var].append(out_dict[var.split('[')[0]][idx1, idx2])
            else:
                if var.startswith('x['):
                    idx = int(var[2:-1])
                elif var.startswith('constraints['):
                    idx = int(var[12:-1])
                elif var.startswith('gradient['):
                    idx = int(var[9:-1])
                elif var.startswith('multipliers['):
                    idx = int(var[12:-1])
                self.var_dict[var].append(out_dict[var.split('[')[0]][idx])
            
            self.lines_dict[var].set_data(x_data, self.var_dict[var])
                
            # Rescale the plot
            self.axs[k].relim()
            self.axs[k].autoscale_view()

        # time.sleep(0.5)

        # Redraw the plot
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()

        self.vis_time += time.time() - v_start
        
    def save_plot(self, save_figname):
        '''
        Save the plot to a file.
        '''
        v_start = time.time()
        # plt.gcf().set_size_inches(10, 3*len(self.visualize_vars))
        # self.fig.set_size_inches(10, 3*len(self.visualize_vars), forward=True)
        self.fig.savefig(save_figname,)
        self.vis_time += time.time() - v_start

    def close_plot(self):
        '''
        Close the plot.
        '''
        self.save_plot(self.save_figname)
        
        plt.ioff()   
        plt.close()

    def keep_plot(self):
        '''
        Keep the plot open after the optimization is completed.
        '''
        self.save_plot(self.save_figname)

        w_start = time.time()
        plt.ioff()
        plt.show()
        self.wait_time += time.time() - w_start


def visualize(savefilename, visualize_vars, itr_start=0, itr_end=-1, major_only=False, save_figname=None):
    '''
    Visualize different scalar variables using the saved data in a file.

    The variables to visualize should be a list of strings, where each string is the name of a variable to visualize. 
    The variables can be any of the following:
    
        - 'objective'       : the objective function value
        - 'optimality'      : the optimality condition
        - 'feasibility'     : the feasibility condition
        - 'x[i]'            : the ith variable value
        - 'constraints[i]'  : the ith constraint value
        - 'jacobian[i,j]'   : the (i,j) element of the Jacobian matrix
        - 'gradient[i]'     : the ith gradient value
        - 'multipliers[i]'  : the ith Lagrange multiplier value

    Creates a plot with the specified variables on the y-axis and the iteration number on the x-axis.
    The plots are stacked vertically in the order they are specified in the list.

    Parameters
    ----------
    savefilename : str
        Path to the saved file.
    visualize_vars : str or list of str
        List of variables to visualize.
    itr_start : int, default=0
        Starting iteration to visualize.
        Negative indices are allowed with -1 representing the last iteration
        and -2 representing the second last iteration and so on.
    itr_end : int, default=-1
        Ending iteration to visualize.
        Negative indices are allowed with -1 representing the last iteration
        and -2 representing the second last iteration and so on.
    major_only : bool, default=False
        If True, only major iterations are visualized.
        If False, all iterations are visualized irrespective of major or line search iterations.
    save_figname : str, default=None
        Path to save the figure. If None, the figure is not saved.

    Examples
    --------
    >>> import numpy as np
    >>> from pyslsqp import optimize
    >>> obj = lambda x: np.sum(x**2)
    >>> grad = lambda x: 2*x
    >>> xl = 0.0
    >>> xu = np.array([1, 1])
    >>> x0 = np.array([0.5, 0.5])
    >>> results = optimize(x0, obj=obj, grad=grad, xl=xl, xu=xu, save_itr='major', save_vars=['objective', 'optimality', 'x'])  # doctest: +ELLIPSIS
    No constraints defined. Running an unconstrained optimization problem...
    Optimization terminated successfully    (Exit mode 0)
                Final objective value                : 0.000000e+00
                Final optimality                     : 0.000000e+00
                Final feasibility                    : 0.000000e+00
                Number of major iterations           : 2
                Number of function evaluations       : 2
                Number of derivative evaluations     : 2
                Average Derivative evaluation time   : ... s per evaluation
                Average Function evaluation time     : ... s per evaluation
                Total Function evaluation time       : ... s [ ...%]
                Total Derivative evaluation time     : ... s [ ...%]
                Optimizer time                       : ... s [ ...%]
                Processing time                      : ... s [ ...%]
                Visualization time                   : ... s [  0.00%]
                Total optimization time              : ... s [100.00%]
                Summary saved to                     : slsqp_summary.out
                Iteration data saved to              : slsqp_recorder.hdf5
    >>> from pyslsqp.postprocessing import visualize
    >>> visualize('slsqp_recorder.hdf5', ['objective', 'optimality', 'x[0]', 'x[1]'], major_only=True)
    '''

    v_start = time.time()
    if plt is None:
        raise ImportError("matplotlib not found, cannot visualize.")
    from pyslsqp.postprocessing import load_variables
    if isinstance(visualize_vars, str):
        visualize_vars = [visualize_vars]
    var_dict = load_variables(savefilename, visualize_vars, itr_start=itr_start, itr_end=itr_end, major_only=major_only)
    
    x_data = np.arange(len(var_dict[visualize_vars[0]]))
    n_plots = len(visualize_vars)
    fig, axs = plt.subplots(n_plots, figsize=(10, 3*n_plots))
    fig.suptitle(f'SLSQP Optimization [{savefilename}]')
    for ax, var in zip(axs, visualize_vars):
        # ax.set_title(var)
        # ax.set_xlabel('Iteration')
        ax.set_ylabel(var)
        if var in ['optimality', 'feasibility']:
            ax.semilogy(x_data, var_dict[var], label=var)
        else:
            ax.plot(x_data, var_dict[var], label=var)

        ax.legend()

    fig.set_size_inches(10, 3*n_plots)
    fig.tight_layout(pad=3.0, h_pad=1, w_pad=1, rect=[0, 0, 1., 1.])
    if save_figname is not None:
        fig.savefig(save_figname)
    plt.show()
    vis_time = time.time() - v_start

if __name__ == '__main__':
    import doctest
    doctest.testmod()


'''
This script tests the optimize function in the pyslsqp module.
'''

def test_get_default_options():
    import numpy as np
    from pyslsqp import get_default_options
    options = get_default_options()
    
    assert options == {
                        'obj': None,
                        'grad': None,
                        'con': None,
                        'jac': None,
                        'meq': 0,
                        'callback': None,
                        'xl': None,
                        'xu': None,
                        'x_scaler': 1.0,
                        'obj_scaler': 1.0,
                        'con_scaler': 1.0,
                        'maxiter': 100,
                        'acc': 1.0E-6,
                        'iprint': 1,
                        'finite_diff_abs_step': None,
                        'finite_diff_rel_step': np.sqrt(np.finfo(float).eps),
                        'summary_filename': 'slsqp_summary.out',
                        'warm_start': False,
                        'hot_start': False,
                        'load_filename': None,
                        'save_itr': None,
                        'save_filename': 'slsqp_recorder.hdf5',
                        'save_vars': ['x', 'objective', 'optimality', 'feasibility', 'step', 'iter', 'majiter', 'ismajor', 'mode'],
                        'visualize': False,
                        'visualize_vars': ['objective', 'optimality', 'feasibility'],
                        'keep_plot_open': False,
                        'save_figname': 'slsqp_plot.pdf',
                    }
    
def test_optimize():
    import numpy as np
    from pyslsqp import optimize
    from numpy.testing import assert_almost_equal
    from testing_utils import check_timing

    def obj(x):
        return np.sum(x**2)
    def grad(x):
        return 2*x

    x0 = np.ones(10)

    # Unconstrained problem and summary file
    res1 = optimize(x0, obj, grad=grad, summary_filename='uncon_slsqp.out')
    res2 = optimize(x0, obj, summary_filename='uncon_fd_slsqp.out')

    assert res1['success'] == True
    assert res2['success'] == True
    assert res1['summary_filename'] == 'uncon_slsqp.out'
    assert res2['summary_filename'] == 'uncon_fd_slsqp.out'
    assert res1['x'].shape == (10,)
    assert res2['x'].shape == (10,)
    assert res1['gradient'].shape == (10,)
    assert res2['gradient'].shape == (10,)
    assert res1['num_majiter'] == 2
    assert res2['num_majiter'] == 2
    assert res1['nfev'] == 3
    assert res2['nfev'] == 3
    assert res1['ngev'] == 2
    assert res2['ngev'] == 2

    check_timing(res1)
    check_timing(res2)

    assert_almost_equal(res1['objective'], 0., decimal=11)
    assert_almost_equal(res1['optimality'], 0., decimal=11)
    assert_almost_equal(res1['feasibility'], 0., decimal=11)
    assert_almost_equal(res1['x'], np.zeros(10), decimal=11)
    assert_almost_equal(res1['gradient'], np.zeros(10), decimal=11)

    assert_almost_equal(res2['objective'], 0., decimal=11)
    assert_almost_equal(res2['optimality'], 0., decimal=11)
    assert_almost_equal(res2['feasibility'], 0., decimal=11)
    assert_almost_equal(res2['x'], np.zeros(10), decimal=11)
    assert_almost_equal(res2['gradient'], np.ones(10)*1.5e-8, decimal=7)

    # Feasibility problem with scalar scaling, and scalar bounds
    def con(x):
        return x[:4] - 0.4
    def jac(x):
        return np.eye(4, 10)
    
    x0 = -np.ones(10)*100
    res3 = optimize(x0, con=con, jac=jac, xl=0.2, xu =0.9, 
                    summary_filename='feas_slsqp.out', x_scaler=0.02, obj_scaler=100.)
    res4 = optimize(x0, con=con, xl=0.2, xu =0.9, 
                    summary_filename='feas_fd_slsqp.out', x_scaler=0.02, obj_scaler=100.)

    assert res3['success'] == True
    assert res4['success'] == True
    assert res3['summary_filename'] == 'feas_slsqp.out'
    assert res4['summary_filename'] == 'feas_fd_slsqp.out'
    
    assert_almost_equal(res3['objective'], 0.0, decimal=11)
    assert_almost_equal(res3['optimality'], 0.0, decimal=11)
    assert_almost_equal(res3['feasibility'], 0.0, decimal=11)
    assert_almost_equal(res3['x'], [0.4, 0.4, 0.4, 0.4, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=11)

    assert_almost_equal(res4['objective'], 0.0, decimal=11)
    assert_almost_equal(res4['optimality'], 0.0, decimal=11)
    assert_almost_equal(res4['feasibility'], 0.0, decimal=11)
    assert_almost_equal(res4['x'], [0.4, 0.4, 0.4, 0.4, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=11)


    x0 = np.ones(10)

    # Bound-constrained problem with vector bounds and vector scaling
    res5 = optimize(x0, obj, grad=grad, xl=0.2*np.ones(10), xu =0.9*np.ones(10), 
                    summary_filename='bdcon_slsqp.out', x_scaler=0.02*np.ones(10), obj_scaler=100.)
    res6 = optimize(x0, obj, xl=0.2*np.ones(10), xu =0.9*np.ones(10), 
                    summary_filename='bdcon_fd_slsqp.out', x_scaler=0.02*np.ones(10), obj_scaler=100.)

    assert res5['success'] == True
    assert res6['success'] == True
    
    assert_almost_equal(res5['objective'], 0.4, decimal=7)
    assert_almost_equal(res5['optimality'], 4e-9, decimal=7)
    assert_almost_equal(res5['feasibility'], 0.0, decimal=11)
    assert_almost_equal(res5['x'], [0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=7)

    assert_almost_equal(res6['objective'], 0.4, decimal=7)
    assert_almost_equal(res6['optimality'], 0.0, decimal=7)
    assert_almost_equal(res6['feasibility'], 0.0, decimal=11)
    assert_almost_equal(res6['x'], [0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=7)

    # Inequality constrained problem with higher accuracy, saving variables and vector/scalar scaling    
    res7 = optimize(x0, obj, grad=grad, con=con, jac=jac, acc=1.0E-10,
                    summary_filename='ineq_slsqp.out', 
                    save_filename='ineq_slsqp.hdf5', save_itr='major', save_vars=['x', 'objective', 'majiter'], 
                    x_scaler=0.02, obj_scaler=100., con_scaler=0.02)
    res8 = optimize(x0, obj, con=con, acc=1.0E-10,
                    summary_filename='ineq_fd_slsqp.out', 
                    save_filename='ineq_fd_slsqp.hdf5', save_itr='major', save_vars=['x', 'objective', 'majiter'], 
                    x_scaler=0.02*np.ones(10), obj_scaler=100., con_scaler=0.02*np.ones(4))
    
    assert res7['success'] == True
    assert res8['success'] == True
    assert res7['save_filename'] == 'ineq_slsqp.hdf5'
    assert res8['save_filename'] == 'ineq_fd_slsqp.hdf5'

    assert_almost_equal(res7['objective'], 0.64, decimal=10)
    assert_almost_equal(res7['optimality'], 0.0, decimal=10)
    assert_almost_equal(res7['feasibility'], 0.0, decimal=10)
    assert_almost_equal(res7['x'], [0.4, 0.4, 0.4, 0.4, 0., 0., 0., 0., 0., 0.], decimal=7)

    assert_almost_equal(res8['objective'], 0.64, decimal=10)
    assert_almost_equal(res8['optimality'], 0.0, decimal=10)
    assert_almost_equal(res8['feasibility'], 0.0, decimal=10)
    assert_almost_equal(res8['x'], [0.4, 0.4, 0.4, 0.4, 0., 0., 0., 0., 0., 0.], decimal=7)


    # Equality constrained problem
    def coneq(x):
        return x[:5] - 0.5
    def jaceq(x):
        return np.eye(5, 10)
    
    res9 = optimize(x0, obj, grad=grad, con=coneq, jac=jaceq, meq=5, 
                    summary_filename='eq_slsqp.out', save_itr='all', save_filename='eq_slsqp.hdf5', 
                    x_scaler=0.02, obj_scaler=100., con_scaler=0.02)
    res10 = optimize(x0, obj, con=coneq, meq=5, 
                     summary_filename='eq_fd_slsqp.out', save_itr='all', save_filename='eq_fd_slsqp.hdf5', 
                     x_scaler=0.02*np.ones(10,), obj_scaler=100., con_scaler=0.02*np.ones(5))
    
    assert res9['success'] == True
    assert res10['success'] == True

    assert_almost_equal(res9['objective'], 1.25, decimal=10)
    assert_almost_equal(res9['optimality'], 1e-8, decimal=7)
    assert_almost_equal(res9['feasibility'], 0.0, decimal=10)
    assert_almost_equal(res9['x'], [0.5, 0.5, 0.5, 0.5, 0.5, 0., 0., 0., 0., 0.], decimal=6)

    assert_almost_equal(res10['objective'], 1.25, decimal=10)
    assert_almost_equal(res10['optimality'], 0.0, decimal=7)
    assert_almost_equal(res10['feasibility'], 0.0, decimal=10)
    assert_almost_equal(res10['x'], [0.5, 0.5, 0.5, 0.5, 0.5, 0., 0., 0., 0., 0.], decimal=6)

    # Equality and inequality constrained problem with saving, vector/scalar scaling and bounds
    def coneqineq(x):
        return np.array([x[0] - 1., x[1] - 3., x[2] - 5.])
    
    def jaceqineq(x):
        return np.eye(3, 10)
    
    res11 = optimize(x0, obj, grad=grad, con=coneqineq, jac=jaceqineq, meq=2, xl=0.2, xu=10., acc=1.0E-6,
                    summary_filename='eqineq_slsqp.out', save_itr='all', save_filename='eqineq_slsqp.hdf5', 
                    x_scaler=0.02, obj_scaler=100., con_scaler=0.02)
    res12 = optimize(x0, obj, con=coneqineq, meq=2, xl=0.2*np.ones(10), xu=10*np.ones(10), acc=1.0E-6,
                    summary_filename='eqineq_fd_slsqp.out', save_itr='all', save_filename='eqineq_fd_slsqp.hdf5', 
                    x_scaler=0.02*np.ones(10), obj_scaler=100., con_scaler=0.02*np.ones(3))
    
    assert res11['success'] == True
    assert res12['success'] == True

    assert_almost_equal(res11['objective'], 35.28, decimal=2)
    assert_almost_equal(res11['optimality'], 0.0, decimal=10)
    assert_almost_equal(res11['feasibility'], 0.0, decimal=3)
    assert_almost_equal(res11['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

    assert_almost_equal(res12['objective'], 35.28, decimal=2)
    assert_almost_equal(res12['optimality'], 0.0, decimal=10)
    assert_almost_equal(res12['feasibility'], 0.0, decimal=7)
    assert_almost_equal(res12['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

import pytest

@pytest.mark.visualize
def test_visualize():
    import os
    import numpy as np
    from numpy.testing import assert_almost_equal
    from pyslsqp import optimize

    # This is the same example as res11 and res12 in test_optimize()

    def obj(x):
        return np.sum(x**2)
    def grad(x):
        return 2*x
      
    def coneqineq(x):
        return np.array([x[0] - 1., x[1] - 3., x[2] - 5.])
    def jaceqineq(x):
        return np.eye(3, 10)
    
    x0 = np.ones(10)

    if os.getenv("GITHUB_ACTIONS") is None: # Skip this test on GitHub Actions since it requires a display
        # This example discovered a bug in SLSQP: majiter jumps from 4 to 9 after 4th major iteration
        # NOTE: <IMPORTANT> This test should not be modified since it exposes a bug in SLSQP, explicit bug fix in pySLSQP.
        # Equality and inequality constrained problem with saving, vector/scalar scaling and bounds
        res1 = optimize(x0, obj, grad=grad, con=coneqineq, jac=jaceqineq, meq=2, xl=0.2, xu=10., acc=1.0E-6,
                        summary_filename='vis_eqineq_slsqp.out', save_itr='all', save_filename='vis_eqineq_slsqp.hdf5', 
                        x_scaler=0.02, obj_scaler=100., con_scaler=0.02, keep_plot_open=True,
                        visualize=True, visualize_vars=['objective', 'optimality', 'feasibility', 'x[0]', 'constraints[0]', 'gradient[0]', 'multipliers[0]', 'multipliers[2]', 'jacobian[0,0]'])
        res2 = optimize(x0, obj, con=coneqineq, meq=2, xl=0.2*np.ones(10), xu=10*np.ones(10), acc=1.0E-6,
                        summary_filename='vis_eqineq_fd_slsqp.out', save_itr='all', save_filename='vis_eqineq_fd_slsqp.hdf5', 
                        x_scaler=0.02*np.ones(10), obj_scaler=100., con_scaler=0.02*np.ones(3), keep_plot_open=True,
                        visualize=True, visualize_vars=['objective', 'optimality', 'feasibility', 'x[0]', 'constraints[0]', 'gradient[0]', 'multipliers[0]', 'multipliers[2]', 'jacobian[0,0]'])
        
        assert res1['success'] == True
        assert res2['success'] == True

        assert_almost_equal(res1['objective'], 35.28, decimal=2)
        assert_almost_equal(res1['optimality'], 0.0, decimal=10)
        assert_almost_equal(res1['feasibility'], 0.0, decimal=3)
        assert_almost_equal(res1['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

        assert_almost_equal(res2['objective'], 35.28, decimal=2)
        assert_almost_equal(res2['optimality'], 0.0, decimal=10)
        assert_almost_equal(res2['feasibility'], 0.0, decimal=7)
        assert_almost_equal(res2['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

def test_warm_and_hot_start():
    import numpy as np
    from numpy.testing import assert_almost_equal
    from pyslsqp import optimize

    # This is the same example as res11 and res12 in test_optimize()

    def obj(x):
        return np.sum(x**2)
    def grad(x):
        return 2*x
      
    def coneqineq(x):
        return np.array([x[0] - 1., x[1] - 3., x[2] - 5.])
    def jaceqineq(x):
        return np.eye(3, 10)
    
    x0 = np.ones(10)

    # Equality and inequality constrained problem with saving, vector/scalar scaling and bounds
    res1 = optimize(x0, obj, grad=grad, con=coneqineq, jac=jaceqineq, meq=2, xl=0.2, xu=10., acc=1.0E-6,
                    summary_filename='save_eqineq_slsqp.out', save_itr='all', save_filename='save_eqineq_slsqp.hdf5',
                    save_vars=['x', 'objective', 'constraints', 'gradient', 'jacobian'],
                    x_scaler=0.02, obj_scaler=100., con_scaler=0.02)
    
    assert res1['success'] == True
    assert_almost_equal(res1['objective'], 35.28, decimal=2)
    assert_almost_equal(res1['optimality'], 0.0, decimal=10)
    assert_almost_equal(res1['feasibility'], 0.0, decimal=3)
    assert_almost_equal(res1['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

    assert res1['num_majiter'] == 5
    assert res1['nfev'] == 6
    assert res1['ngev'] == 5
    
    # Warm start
    res2 = optimize(x0, obj, grad=grad, con=coneqineq, jac=jaceqineq, meq=2, xl=0.2, xu=10., acc=1.0E-6,
                    summary_filename='warm_eqineq_slsqp.out', warm_start=True, load_filename='save_eqineq_slsqp.hdf5', 
                    save_itr='all', save_filename='save_eqineq_slsqp.hdf5', x_scaler=0.02, obj_scaler=100., con_scaler=0.02)
    
    assert res2['success'] == True
    assert res2['save_filename'] == 'save_eqineq_slsqp_warm.hdf5'
    assert_almost_equal(res2['objective'], 35.28, decimal=2)
    assert_almost_equal(res2['optimality'], 0.0, decimal=10)
    assert_almost_equal(res2['feasibility'], 0.0, decimal=3)
    assert_almost_equal(res2['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

    assert res2['num_majiter'] == 1
    assert res2['nfev'] == 1
    assert res2['ngev'] == 1

    # Hot start
    res3 = optimize(x0, obj, grad=grad, con=coneqineq, jac=jaceqineq, meq=2, xl=0.2, xu=10., acc=1.0E-6,
                    summary_filename='hot_eqineq_slsqp.out', hot_start=True, load_filename='save_eqineq_slsqp.hdf5',
                    save_itr='all', save_filename='hot_eqineq_slsqp.hdf5', x_scaler=0.02, obj_scaler=100., con_scaler=0.02)
    
    assert res3['success'] == True
    assert res3['save_filename'] == 'hot_eqineq_slsqp.hdf5'
    assert_almost_equal(res3['objective'], 35.28, decimal=2)
    assert_almost_equal(res3['optimality'], 0.0, decimal=10)
    assert_almost_equal(res3['feasibility'], 0.0, decimal=3)
    assert_almost_equal(res3['x'], [1., 3., 5., 0.2, 0.2, 0.2, 0.2, 0.2, 0.2, 0.2], decimal=3)

    assert res3['num_majiter'] == 5
    assert res3['nfev'] == 6
    assert res3['ngev'] == 5
    assert res3['nfev_reused_in_hotstart'] == 6
    assert res3['ngev_reused_in_hotstart'] == 5

    
if __name__ == "__main__":
    test_optimize()
    test_visualize()
    test_get_default_options()
    test_warm_and_hot_start()